#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
#
# This script is licensed under GNU GPL version 2.0 or above
# (c) 2022 Antonio J. Delgado
# Export to node exporter the summary of a restic backup

import sys
import os
import logging
import json
from logging.handlers import SysLogHandler
import time
from datetime import datetime, timedelta
import click
import click_config_file
from systemd import journal

class ResticExporter:
    '''Exporter of Restic data to Prometheus'''

    def __init__(self, **kwargs):
        ''' Initial function called when object is created '''
        self.config = kwargs
        if 'log_file' not in kwargs or kwargs['log_file'] is None:
            self.config['log_file'] = os.path.join(
                os.environ.get(
                    'HOME',
                    os.environ.get(
                        'USERPROFILE',
                        os.getcwd()
                    )
                ),
                'log',
                '__project_codename__.log'
            )
        self._init_log()

        self.metric_name = self.config['metric_name'].replace(' ', '_')
        self.job_name = self.config['job_name'].replace(' ', '_')
        self.metric_description = self.config['metric_description']
        self.labels= {"job_name": self.config['job_name']}

        if self.config['extra_labels']:
            self.labels = {
                **self.labels,
                **self._read_extra_labels(self.config['extra_labels'])
            }
        if self.config['systemd']:
            self._read_summary_from_json(self.config['json_file'])
        else:
            self._read_summary_from_systemd()

        self._write_metrics()

    def _read_summary_from_systemd(self,):
        self.summaries = []
        journal_obj = journal.Reader()
        journal_obj.seek_realtime(datetime.now() - timedelta(days=1))
        journal_obj.add_match('_EXE=/usr/local/bin/restic')
        for entry in journal_obj:
            if '"message_type":"summary"' in entry['MESSAGE']:
                summary = json.loads(entry['MESSAGE'])
                summary['metric_name'] = 'restic_summary'
                summary['job_name'] = entry['_SYSTEMD_UNIT'].replace(
                    'remote_backupjob_',
                    '').replace('.service', '')
                self.summaries.append(summary)

    def _write_metrics(self):
        for summary in self.summaries:
            labels = self._convert_labels(self.labels)
            counters = [
                'files_new',
                'files_changed',
                'files_unmodified',
                'dirs_new',
                'dirs_changed',
                'dirs_unmodified',
                'data_blobs',
                'tree_blobs',
                'data_added',
                'total_files_processed',
                'total_bytes_processed',
                'total_duration'
            ]
            if self.config['output_file'] != "-":
                outfile = open(self.config['output_file'], 'w', encoding='utf-8')
            else:
                outfile = sys.stdout
            outfile.write(
                f"# HELP {summary['metric_name']}_{summary['job_name']} \
{self.metric_description}.\n"
            )
            outfile.write(f"# TYPE {summary['metric_name']}_{summary['job_name']} counter\n")
            for counter in counters:
                if counter in summary:
                    outfile.write(
                        f"{summary['metric_name']}_{counter}{labels} {float(summary[counter])}\n"
                    )

    def _read_summary_from_json(self, json_file):
        try:
            with open(json_file, 'r', encoding='utf-8') as file_pointer:
                content = file_pointer.readlines()
        except Exception as error:
            self._log.error(
                "# Error reading file '%s'. Check permissions. %s",
                json_file,
                error
            )

        summary = {
            "timestamp": time.time(),
            "metric_name": self.metric_name,
            "job_name": self.job_name,
        }
        for line in content:
            try:
                line_data = json.loads(line)
                if 'message_type' in line_data:
                    if line_data['message_type'] == 'summary':
                        self.summary = line_data
                        if 'snapshot_id' in line_data:
                            self.labels["snapshot_id"] = line_data['snapshot_id']
            except json.decoder.JSONDecodeError as error:
                fixed_line = line.replace('\n', '')
                self._log.error(
                    "# Error decoding line '%s'. %s",
                    fixed_line,
                    error
                )
        file_stats = os.stat(json_file)
        summary['timestamp'] = round(file_stats.st_mtime * 1000)
        self._log.debug("# Summary: {json.dumps(summary, indent=2)}")
        self._log.debug(
            "# Labels: %s",
            self.labels
        )
        self.summaries = [ summary ]

    def _read_extra_labels(self, extra_labels):
        labels_ls = {}
        for pair in extra_labels.split(','):
            if '=' in pair:
                key, value=pair.split('=', 1)
                labels_ls[key] = value
                self._log.debug(
                    "# Added extra label '%s'='%s'",
                    key,
                    value
                )
        return labels_ls

    def _convert_labels(self, labels):
        labels_ls = []
        for key in labels.keys():
            labels_ls.append(f"{key}=\"{labels[key]}\"")
        text_labels = ','.join(labels_ls)
        labels_string = "{" + text_labels + "}"
        self._log.debug(
            "# Labels: %s",
            labels_string
        )
        return labels_string

    def _init_log(self):
        ''' Initialize log object '''
        self._log = logging.getLogger("restic_exporter")
        self._log.setLevel(logging.DEBUG)

        sysloghandler = SysLogHandler()
        sysloghandler.setLevel(logging.DEBUG)
        self._log.addHandler(sysloghandler)

        streamhandler = logging.StreamHandler(sys.stdout)
        streamhandler.setLevel(logging.getLevelName(self.config.get("debug_level", 'INFO')))
        self._log.addHandler(streamhandler)

        if 'log_file' in self.config:
            log_file = self.config['log_file']
        else:
            home_folder = os.environ.get('HOME', os.environ.get('USERPROFILE', ''))
            log_folder = os.path.join(home_folder, "log")
            log_file = os.path.join(log_folder, "restic_exporter.log")

        if not os.path.exists(os.path.dirname(log_file)):
            os.mkdir(os.path.dirname(log_file))

        filehandler = logging.handlers.RotatingFileHandler(log_file, maxBytes=102400000)
        # create formatter
        formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
        filehandler.setFormatter(formatter)
        filehandler.setLevel(logging.DEBUG)
        self._log.addHandler(filehandler)
        return True

@click.command()
@click.option("--debug-level", "-d", default="INFO",
    type=click.Choice(
        ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
        case_sensitive=False,
    ), help='Set the debug level for the standard output.')
@click.option('--log-file', '-l', help="File to store all debug messages.")
@click.option("--json-file", "-j", help='JSON file containing the output of restic')
@click.option(
    '--job-name', '-n',
    required=True,
    help='Restic job name to attach to the exported metrics'
)
@click.option(
    '--extra-labels', '-a',
    required=False,
    default=None,
    help='Pairs key=value separated by commas with extra labels to add to the summary'
)
@click.option(
    '--metric-name', '-m',
    default='restic_summary',
    help='Metric name. Spaces will be replaced with underscore (_).'
)
@click.option(
    '--metric-description', '-d',
    default='Restic summary metrics',
    help='Metric description.'
)
@click.option(
    '--systemd', '-s',
    default=False,
    help='Get JSON data from Systemd units',
)
@click.option(
    '--output-file',
    '-o',
    default='-',
    help='File to write the metrics so node-exporter can read it. Use "-" \
to write to standard output (default).'
)
@click_config_file.configuration_option()
def __main__(**kwargs):
    return ResticExporter(**kwargs)

if __name__ == "__main__":
    __main__()
